/*
 *  Colditz Escape - Rewritten Engine for "Escape From Colditz"
 *  copyright (C) 2008-2017 Aperture Software
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  ---------------------------------------------------------------------------
 *  soundplayer.cpp: SFX and MOD Player Implementation
 *  Copied almost 100% from PSP SDK's "PSP ModPlayer v1.0" by adresd:
 *     http://svn.pspdev.org/listing.php?repname=pspware&path=%2Ftrunk%2FPSPMediaCenter%2Fcodec%2Fmod%2F&rev=0&sc=0
 *  ---------------------------------------------------------------------------
 */

#if defined(WIN32)
#define _WIN32_DCOM
#pragma warning(disable:4995)
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "low-level.h"
#include "audio_backend.h"
#include "soundplayer.h"
#include "soundtables.h"

#if !defined(min)
#define min(a,b) (((a)<(b))?(a):(b))
#endif
#if !defined(max)
#define max(a,b) (((a)>(b))?(a):(b))
#endif

// As a sample is being mixed into the buffer its position pointer is updated.
// The position pointer is a 32-bit integer that is used to store a fixed-point
// number. The following constant specifies how many of the bits should be used
// for the fractional part.
#define FRAC_BITS 10

// Maximum number of channels we'll handle
#define NB_CHANNELS 4

#define AMIGA_VOLUME_MAX 64


//////////////////////////////////////////////////////////////////////
// Function Prototypes
//////////////////////////////////////////////////////////////////////
static void SetMasterVolume(int volume);
static int  ReadModWord(unsigned char *data, int index);
static void DoTremolo(int track);
static void DoVibrato(int track);
static void DoPorta(int track);
static void SlideVolume(int track, int amount);
static int  VolumeTable[65][256];
static void UpdateEffects();
static void UpdateRow();
static bool MixChunk(int numsamples, short *buffer);
static bool MixSubChunk(short *buffer, int numsamples);
static void ModPlayCallback(void *_buf2, unsigned int length, void *pdata);
static void LoopCallback(void *_buf2, unsigned int length, void *pdata);

//////////////////////////////////////////////////////////////////////
// Global local variables
//////////////////////////////////////////////////////////////////////

// The following variables contain the music data, ie they don't change value until you load a new file
static char m_szName[200];
static int m_nSongLength;

static int m_nOrders_num;
static int *m_nOrders;
static int m_Patterns_num;
static Pattern *m_Patterns = NULL;
static int m_Samples_num;
static Sample *m_Samples;

// The following variables are maintained and updated by the tracker during playback
static int m_nSpeed;			// Speed of mod being played
static int m_nOrder;			// Current order being played
static int m_nRow;				// Current row being played
static int m_nTick;				// Current tick number (there are "m_nSpeed"
								// ticks between each row)
static int BPM_RATE;			// Adjusted master BPM for refresh rate
static int m_nBPM;				// Beats-per-minute...controls length of
								// each tick
static int m_nSamplesLeft;		// Number of samples left to mix for the
								// current tick
static int m_nNumTracks;		// Number of tracks in this mod

static int m_TrackDat_num;
static TrackData *m_TrackDat;	// Stores info for each track being played
static RowData *m_CurrentRow;	// Pointer to the current row being played
static bool m_bPlaying = false;	// Set to true when a mod is being played
static bool m_bSet     = false;	// True when a mod has been set
static bool no_audio   = true;	// Is the audio working
static uint8_t *data;

//////////////////////////////////////////////////////////////////////
// These are the public functions
//////////////////////////////////////////////////////////////////////
// This could be used to set a different channel/voice for mod playout
static int mod_channel = 0;
static int sfx_channel = 0;
// for loops
static int				loop_channel = -1;
static unsigned long	loop_pos;
static unsigned long	loop_len;
static void*			loop_addr;

// Play a MONO, possibly looping, sound sample. channel = -1 => autoallocated
// NB: the PSP is limited to 64 KB for non looping samples
bool play_sample(int channel, unsigned int volume, void *address, unsigned int length,
                 unsigned int frequency, unsigned int bits_per_sample, bool loop)
{
    int play_channel;

    if (no_audio)
        return false;

    if (channel == -1)
    {
        play_channel = sfx_channel;
        sfx_channel = (sfx_channel+1)%NB_CHANNELS;
    }
    else
        play_channel = channel;

    // We might have a loop currently playing on that channel
    if (play_channel == loop_channel)
        stop_loop();

    if (loop)
    {	// Play a looping sample
        loop_pos = 0;
        loop_len = length;
        loop_addr = address;
        loop_channel = play_channel;
    }

#if defined(PSP)
    // We upconverted the samples to 16-bit for PSP usage
    bits_per_sample = 16;
#endif

    audio_backend_release_voice(play_channel);
    // Our SFXs are always mono
    if (loop)
    {
        if (!audio_backend_set_voice_callback(loop_channel, LoopCallback, NULL, frequency, bits_per_sample, false))
            return false;
    }
    else
    {
        if (!audio_backend_set_voice(play_channel, address, length, frequency, bits_per_sample, false))
            return false;
    }
    audio_backend_set_voice_volume(play_channel, (float) volume/AMIGA_VOLUME_MAX);
    audio_backend_start_voice(play_channel);

    return true;
}

// Someone might get tired of looping SFXs
void stop_loop()
{
    if (no_audio)
        return;

    if (loop_channel == -1)
        return;

    audio_backend_stop_voice(loop_channel);
    audio_backend_release_voice(loop_channel);
    loop_channel = -1;
}

#if defined(PSP)
#include <pspaudio.h>   // PSP_AUDIO_SAMPLE_ALIGN()
// This upconverts an 8 bit PCM file @ any frequency to 16 bit/PLAYBACK_FREQ
bool psp_upsample(short **upconverted_address, unsigned long *upconverted_length, char *src_sample,
                  unsigned long src_numsamples, unsigned short src_frequency)
{
    unsigned long src_fracpos = 0;
    unsigned long dst_pos = 0;
    unsigned long delta_pos = (int) (src_frequency * (1 << FRAC_BITS) / PLAYBACK_FREQ);
    unsigned long src_fraclength = src_numsamples << FRAC_BITS;
    unsigned long dst_length = (src_fraclength - 1) / delta_pos + 1;
    unsigned long dst_length_aligned = PSP_AUDIO_SAMPLE_ALIGN(dst_length);
    short* dst_sample = (short*) aligned_malloc(2*dst_length_aligned, 64);

    if (dst_sample == NULL)
    {
        fprintf(stderr, "psp_upsample: malloc error\n");
        return false;
    }

    for (dst_pos = 0; dst_pos < dst_length; dst_pos++)
    {
        // Mix this sample in and update our position
    #ifdef OVERSAMPLE
        int sample1 = src_sample[src_fracpos >> FRAC_BITS] << 8;
        int sample2 = src_sample[(src_fracpos >> FRAC_BITS) + 1] << 8;
        int frac1 = src_fracpos & ((1 << FRAC_BITS) - 1);
        int frac2 = (1 << FRAC_BITS) - frac1;
        dst_sample[dst_pos] = ((sample1 * frac2) + (sample2 * frac1)) >> FRAC_BITS;
    #else
        dst_sample[dst_pos] = (src_sample[src_fracpos >> FRAC_BITS]) << 8;
    #endif
        src_fracpos += delta_pos;
    }

    *upconverted_address = dst_sample;
    *upconverted_length = dst_length_aligned;

    return true;
}
#endif

static void ModPlayCallback(void *_buf2, unsigned int length, void *pdata)
{
    short *_buf = (short *)_buf2;
    if (m_bPlaying == true) {	//  Playing , so mix up a buffer
        MixChunk(length, _buf);
    } else {			//  Not Playing , so clear buffer
        {
            unsigned int count;
            for (count = 0; count < length * 2; count++)
                *(_buf + count) = 0;
        }
    }
}

static void LoopCallback(void *_buf2, unsigned int length, void *pdata)
{
register unsigned int i;
#if defined(PSP)
    // 16 bit Stereo only
    short* _addr = (short*)loop_addr;
    short* _buf  = (short*)_buf2;
    for (i=0; i<length; i++)
    {
        // 16 bit Mono -> 16 bit Stereo
        _buf[2*i] = _addr[loop_pos];
        _buf[2*i+1] = _addr[loop_pos];
        loop_pos = (loop_pos+1)%loop_len;
    }
#else
    // 8 bit mono is fine
    unsigned char* _addr = (unsigned char*)loop_addr;
    unsigned char* _buf = (unsigned char*)_buf2;
    for (i=0; i<length; i++)
    {
        _buf[i] = _addr[loop_pos];
        loop_pos = (loop_pos+1)%loop_len;
    }
#endif
}

// Initialize the audio engine
bool audio_init()
{
    if (!audio_release())
        fprintf(stderr, "audio_release error\n");
    if (!audio_backend_init())
        return false;
    no_audio = false;
    return true;
}

// Release the audio engine
bool audio_release()
{
    return audio_backend_release();
}

// Terminate MOD playout
void mod_release()
{
    int i;
    int row;

    if (no_audio)
        return;

    mod_stop();
    audio_backend_release_voice(mod_channel);

    if (!m_bSet)
        return;

    // Tear down all the mallocs done
    //free the file itself
    if (data)
        free(data);
    // Free patterns
    if (m_Patterns != NULL)
        for (i = 0; i < m_Patterns_num; i++) {
            for (row = 0; row < 64; row++)
                free(m_Patterns[i].row[row].note);
            free(m_Patterns[i].row);
        }
    // Free orders
    free(m_nOrders);

    // Free samples
    for (i = 1; i < m_Samples_num; i++)
        free(m_Samples[i].data);
    free(m_Samples);

    // Free tracks and patterns
    free(m_TrackDat);
    free(m_Patterns);

    m_bSet = false;
}

bool is_mod_playing()
{
    return m_bPlaying;
}

//  This is the initialiser and module loader
//  This is a general call, which loads the module from the
//  given address into the modplayer
//
//  It basically loads into an internal format, so once this function
//  has returned the buffer at 'data' will not be needed again.
bool mod_init(char *filename)
{
    int i, numpatterns, row, note, size = 0;
    int modindex = 0;
    int numsamples;
    char modname[21];
    FILE* fd;

    if (no_audio)
        return false;

    m_bPlaying = false;

    if ((fd = fopen (filename, "rb")) != NULL) {
        //  opened file, so get size now
        fseek(fd, 0, SEEK_END);
        size = ftell(fd);
        if (size < 1084) {
            fclose(fd);
            return false;
        }
        fseek(fd, 0, SEEK_SET);
        data = (unsigned char *) malloc(size + 8);
        if (data != NULL) {	// Read file in
            memset(data, 0, size + 8);
            IGNORE_RETVAL(fread(data, 1, size, fd));
        } else {
            fclose(fd);
            return false;
        }
        // Close file
        fclose(fd);
    } else 	// we couldn't open the file
        return false;

    // Release the voice first
    audio_backend_release_voice(0);
    if (!audio_backend_set_voice_callback(mod_channel, ModPlayCallback, NULL, PLAYBACK_FREQ, 16, true))
        return false;

    BPM_RATE = 130;
    //  Set default settings
    numsamples = 32;
    m_nNumTracks = 4;
    //  Check for diff types of mod
    if ((data[1080] == 'M') && (data[1081] == '.') && (data[1082] == 'K') && (data[1083] == '.'));
    else if ((data[1080] == 'F') && (data[1081] == 'L') && (data[1082] == 'T') && (data[1083] == '4'));
    else if ((data[1080] == 'F') && (data[1081] == 'L') && (data[1082] == 'T') && (data[1083] == '8'))
        m_nNumTracks = 8;
    else if ((data[1080] == '6') && (data[1081] == 'C') && (data[1082] == 'H') && (data[1083] == 'N'))
        m_nNumTracks = 6;
    else if ((data[1080] == '8') && (data[1081] == 'C') && (data[1082] == 'H') && (data[1083] == 'N'))
        m_nNumTracks = 8;
    else
        numsamples = 16;

    //  Setup the trackdata structure
    m_TrackDat_num = m_nNumTracks;
    m_TrackDat = (TrackData *) malloc(m_TrackDat_num * sizeof(TrackData));

    // Get the name
    memcpy(modname, &data[modindex], 20);
    modname[20] = 0;
    strcpy(m_szName, modname);
    modindex += 20;

    // Read in all the instrument headers - mod files have 31, sample #0 is ignored
    m_Samples_num = numsamples;
    // Make sure we allocate zeroed data
    m_Samples = (Sample *) calloc(m_Samples_num,  sizeof(Sample));
    if (m_Samples == NULL)
        return false;
    // BUGFIX: If you don't keep the first Sample to 0,
    // all kind of bad things can happen on Win32!!!
    for (i = 1; i < numsamples; i++) {
        // Read the sample name
        memcpy(m_Samples[i].szName, &data[modindex], 22);
        modindex += 22;
        // Read remaining info about sample
        m_Samples[i].nLength = ReadModWord(data, modindex);
        modindex += 2;
        m_Samples[i].nFineTune = (int) (unsigned char) *(data + modindex);
        modindex++;
        if (m_Samples[i].nFineTune > 7)
            m_Samples[i].nFineTune -= 16;
        m_Samples[i].nVolume = (int) (unsigned char) *(data + modindex);
        modindex++;
        m_Samples[i].nLoopStart = ReadModWord(data, modindex);
        modindex += 2;
        m_Samples[i].nLoopLength = ReadModWord(data, modindex);
        modindex += 2;
        m_Samples[i].nLoopEnd = m_Samples[i].nLoopStart + m_Samples[i].nLoopLength;

        // Fix loop end in case it goes too far
        if (m_Samples[i].nLoopEnd > m_Samples[i].nLength)
            m_Samples[i].nLoopEnd = m_Samples[i].nLength;
    }

    // Read in song data
    m_nSongLength = (int) (unsigned char) *(data + modindex);
    modindex++;
    modindex++;			// Skip over this byte, it's no longer used

    numpatterns = 0;
    m_nOrders_num = 128;
    m_nOrders = (int *) malloc(m_nOrders_num * sizeof(int));
    if (m_nOrders == NULL)
        return false;
    for (i = 0; i < 128; i++) {
        m_nOrders[i] = (int) (unsigned char) *(data + modindex);
        modindex++;
        if (m_nOrders[i] > numpatterns)
            numpatterns = m_nOrders[i];
    }
    numpatterns++;
    modindex += 4;			// skip over the identifier

    // Load in the pattern data
    m_Patterns_num = numpatterns;
    m_Patterns = (Pattern *) malloc(m_Patterns_num * sizeof(Pattern));
    if (m_Patterns == NULL)
        return false;
    for (i = 0; i < numpatterns; i++) {
        // Set the number of rows for this pattern, for mods it's always 64
        m_Patterns[i].numrows = 64;
        m_Patterns[i].row = (RowData *) malloc(m_Patterns[i].numrows * sizeof(RowData));

        // Loop through each row
        for (row = 0; row < 64; row++) {
            // Set the number of notes for this pattern
            m_Patterns[i].row[row].numnotes = m_nNumTracks;
            m_Patterns[i].row[row].note = (NoteData *) malloc(m_Patterns[i].row[row].numnotes * sizeof(NoteData));

            // Loop through each note
            for (note = 0; note < m_nNumTracks; note++) {
                int b0, b1, b2, b3, period;
                // Get the 4 bytes for this note
                b0 = (int) (unsigned char) *(data + modindex);
                b1 = (int) (unsigned char) *(data + modindex + 1);
                b2 = (int) (unsigned char) *(data + modindex + 2);
                b3 = (int) (unsigned char) *(data + modindex + 3);
                modindex += 4;

                // Parse them
                period = ((b0 & 0x0F) << 8) | b1;
                if (period)
                    //m_Patterns[i].row[row].note[note].period_index = (int)((log(856) - log(period)) / log(1.007246412224) + 8);// ??
                    m_Patterns[i].row[row].note[note].period_index = Period_Log_Lookup[period];
                else
                    m_Patterns[i].row[row].note[note].period_index = -1;

                m_Patterns[i].row[row].note[note].sample_num = (b0 & 0xF0) | (b2 >> 4);
                m_Patterns[i].row[row].note[note].effect = b2 & 0x0F;
                m_Patterns[i].row[row].note[note].effect_parms = b3;
            }
        }
    }

    // Load in the sample data
    for (i = 1; i < numsamples; i++) {
        int length;
        m_Samples[i].data_length = m_Samples[i].nLength;
        m_Samples[i].data = (char *) malloc(m_Samples[i].data_length + 1);
        memset(m_Samples[i].data, 0, m_Samples[i].data_length + 1);

        if (m_Samples[i].nLength) {
            memcpy(&m_Samples[i].data[0], &data[modindex], m_Samples[i].nLength);
        }
        modindex += m_Samples[i].nLength;

        // Duplicate the last byte, we'll need an extra one in order to safely anti-alias
        length = m_Samples[i].nLength;
        if (length > 0) {
            m_Samples[i].data[length] = m_Samples[i].data[length - 1];

            if (m_Samples[i].nLoopLength > 2)
                m_Samples[i].data[m_Samples[i].nLoopEnd] = m_Samples[i].data[m_Samples[i].nLoopStart];
        }
    }
    //  Set volume to full ready to play
    SetMasterVolume(64);
    m_bPlaying = false;
    m_bSet = true;

    return true;
}

// Starts playing a MOD
bool mod_play()
{
    int track;

    if (no_audio)
        return false;

    // See if I'm already playing
    if (m_bPlaying)
        return false;

    if ((m_Patterns == NULL)|| (m_nOrders == NULL))
        return false;

    // Reset all track data
    for (track = 0; track < m_nNumTracks; track++) {
        m_TrackDat[track].sample = 0;
        m_TrackDat[track].pos = 0;
        m_TrackDat[track].period_index = 0;
        m_TrackDat[track].period = 0;
        m_TrackDat[track].volume = 0;
        m_TrackDat[track].mixvol = 0;
        m_TrackDat[track].porta = 0;
        m_TrackDat[track].portasp = 0;
        m_TrackDat[track].vibspe = 0;
        m_TrackDat[track].vibdep = 0;
        m_TrackDat[track].tremspe = 0;
        m_TrackDat[track].tremdep = 0;
        m_TrackDat[track].panval = 0;
        m_TrackDat[track].freq = 0;
        m_TrackDat[track].sinepos = 0;
        m_TrackDat[track].sineneg = 0;
    }

    // Get ready to play
    m_nSpeed = 6;
    m_nBPM = BPM_RATE;
    m_nSamplesLeft = 0;
    m_nOrder = 0;
    m_nRow = 0;
    m_CurrentRow = &m_Patterns[m_nOrders[m_nOrder]].row[m_nRow];
    m_nTick = 0;

    m_bPlaying = true;
    return audio_backend_start_voice(0);
}

void mod_pause()
{
    m_bPlaying = !m_bPlaying;
}

bool mod_stop()
{
    if (no_audio)
        return false;

    m_bPlaying = false;
    return audio_backend_stop_voice(0);
}

//////////////////////////////////////////////////////////////////////
// Functions - Local and not public
//////////////////////////////////////////////////////////////////////

// This function mixes an entire chunk of sound which is then
// to be sent to the sound driver, in this case the IOP module.
static bool MixChunk(int numsamples, short *buffer)
{
    // Calculate the number of samples per beat
    //  48000 / (125 * 2 / 5) = 48000/ 50 = 960
    int samples_per_beat = PLAYBACK_FREQ / (m_nBPM * 2 / 5);

    int thiscount;
    // Keep looping until we've filled the buffer
    int tickdata = 0;
    int samples_to_mix = numsamples;

    while (samples_to_mix) {
        // Only move on to the next tick if we finished mixing the last
        if (!m_nSamplesLeft) {
            // If we're on tick 0 then update the row
            if (m_nTick == 0) {
                // Get this row
                if (m_nOrders != NULL) {
                    m_CurrentRow = &m_Patterns[m_nOrders[m_nOrder]].row[m_nRow];

                    // Set up for next row (effect might change these values later)
                    m_nRow++;
                    if (m_nRow >= 64) {
                        m_nRow = 0;
                        m_nOrder++;
                        if (m_nOrder >= m_nSongLength)
                            m_nOrder = 0;
                    }
                    // Now update this row
                    UpdateRow();
                }
            }
            // Otherwise, all we gotta do is update the effects
            else {
                UpdateEffects();
            }

            // Move on to next tick
            m_nTick++;
            if (m_nTick >= m_nSpeed)
                m_nTick = 0;

            // Set the number of samples to mix in this chunk
            m_nSamplesLeft = samples_per_beat;
        }
        // Ok, so we know that we gotta mix 'm_nSamplesLeft' samples into
        // this buffer, see how much room we actually got
        thiscount = m_nSamplesLeft;
        if (thiscount > samples_to_mix)
            thiscount = samples_to_mix;

        // Make a note that we've added this amount
        m_nSamplesLeft -= thiscount;
        samples_to_mix -= thiscount;

        // Now mix it!
        MixSubChunk(&buffer[tickdata * 2], thiscount);
        tickdata += thiscount;
    }
    return true;
}

// This function is called whenever a new row is encountered.
// It loops through each track, check's it's appropriate NoteData structure
// and updates the track accordingly.
static void UpdateRow()
{
    int neworder = m_nOrder;
    int newrow = m_nRow;
    int track;

    // Loop through each track
    for (track = 0; track < m_nNumTracks; track++) {
        // Get note data
        NoteData *note = &m_CurrentRow->note[track];

        // Make a copy of each value in the NoteData structure so they'll
        // be easier to work with (less typing)
        int sample = note->sample_num;
        int period = note->period_index;
        int effect = note->effect;
        int eparm = note->effect_parms;
        int eparmx = eparm >> 4;	// effect parameter x
        int eparmy = eparm & 0xF;	// effect parameter y

        // Are we changing the sample being played?
        if (sample > 0) {
            m_TrackDat[track].sample = sample;
            m_TrackDat[track].volume = m_Samples[sample].nVolume;
            m_TrackDat[track].mixvol = m_TrackDat[track].volume;
            if ((effect != 3) && (effect != 5))
                m_TrackDat[track].pos = 0;
        }
        // Are we changing the frequency being played?
        if (period >= 0) {
            // Remember the note
            m_TrackDat[track].period_index = note->period_index;

            // If not a porta effect, then set the channels frequency to the
            // looked up amiga value + or - any finetune
            if ((effect != 3) && (effect != 5)) {
                int notenum = m_TrackDat[track].period_index + m_Samples[m_TrackDat[track].sample].nFineTune;
                if (notenum < 0)
                    notenum = 0;
                if (notenum > NUMNOTES - 1)
                    notenum = NUMNOTES - 1;
                m_TrackDat[track].period = PeriodTable[notenum];
            }
            // If there is no sample number or effect then we reset the position
            if ((sample == 0) && (effect == 0))
                m_TrackDat[track].pos = 0;

            // Now reset a few things
            m_TrackDat[track].vibspe = 0;
            m_TrackDat[track].vibdep = 0;
            m_TrackDat[track].tremspe = 0;
            m_TrackDat[track].tremdep = 0;
            m_TrackDat[track].sinepos = 0;
            m_TrackDat[track].sineneg = 0;
        }
        // Process any effects - need to include 1, 2, 3, 4 and A
        switch (note->effect) {
            // Arpeggio
        case 0x00:
            break;		// tick effect

            // Porta Up
        case 0x01:
            break;		// tick effect

            // Porta Down
        case 0x02:
            break;		// tick effect

            // Porta to Note (3) and Porta + Vol Slide (5)
        case 0x03:
        case 0x05:
            m_TrackDat[track].porta = PeriodTable[m_TrackDat[track].period_index + m_Samples[sample].nFineTune];
            if (eparm > 0 && effect == 0x3)
                m_TrackDat[track].portasp = eparm;
            break;

            // Vibrato
        case 0x04:
            if (eparmx > 0)
                m_TrackDat[track].vibspe = eparmx;
            if (eparmy > 0)
                m_TrackDat[track].vibdep = eparmy;
            break;

            // Vibrato + Vol Slide
        case 0x06:
            break;		// tick effect

            // Tremolo
        case 0x07:
            if (eparmx > 0)
                m_TrackDat[track].tremspe = eparmx;
            if (eparmy > 0)
                m_TrackDat[track].tremdep = eparmy;
            break;

            // Pan - not supported in the mixing yet
        case 0x08:
            if (eparm == 0xa4)
                m_TrackDat[track].panval = 7;
            else
                m_TrackDat[track].panval = (eparm >> 3) - 1;
            if (m_TrackDat[track].panval < 0)
                m_TrackDat[track].panval = 0;
            break;

            // Sample offset
        case 0x09:
            m_TrackDat[track].pos = note->effect_parms << (FRAC_BITS + 8);
            break;

            // Volume Slide
        case 0x0A:
            break;		// tick effect

            // Jump To Pattern
        case 0x0B:
            neworder = note->effect_parms;
            if (neworder >= m_nSongLength)
                neworder = 0;
            newrow = 0;

            // Set Volume
        case 0x0C:
            m_TrackDat[track].volume = note->effect_parms;
            SlideVolume(track, 0);
            m_TrackDat[track].mixvol = m_TrackDat[track].volume;
            break;

            // Break from current pattern
        case 0x0D:
            newrow = eparmx * 10 + eparmy;
            if (newrow > 63)
                newrow = 0;
            neworder = m_nOrder + 1;
            if (neworder >= m_nSongLength)
                neworder = 0;
            break;

            // Extended effects
        case 0x0E:
            switch (eparmx) {
                // Set filter
            case 0x00:
                break;		// not supported

                // Fine porta up
            case 0x01:
                m_TrackDat[track].period -= eparmy;
                break;

                // Fine porta down
            case 0x02:
                m_TrackDat[track].period += eparmy;
                break;

                // Glissando
            case 0x03:
                break;		// not supported


                // Set vibrato waveform
            case 0x04:
                break;		// not supported

                // Set finetune
            case 0x05:
                m_Samples[sample].nFineTune = eparmy;
                if (m_Samples[sample].nFineTune > 7)
                    m_Samples[sample].nFineTune -= 16;
                break;

                // Pattern loop
            case 0x6:
                break;		// not supported

                // Set tremolo waveform
            case 0x07:
                break;		// not supported

                // Pos panning - not supported in the mixing yet
            case 0x08:
                m_TrackDat[track].panval = eparmy;
                break;

                // Retrig Note
            case 0x09:
                break;		// tick effect

                // Fine volside up
            case 0x0A:
                SlideVolume(track, eparmy);
                m_TrackDat[track].mixvol = m_TrackDat[track].volume;
                break;

                // Fine volside down
            case 0xB:
                SlideVolume(track, -eparmy);
                m_TrackDat[track].mixvol = m_TrackDat[track].volume;
                break;

                // Cut note
            case 0x0C:
                break;		// tick effect

                // Delay note
            case 0x0D:
                break;		// not supported

                // Pattern delay
            case 0x0E:
                break;		// not supported

                // Invert loop
            case 0x0F:
                break;		// not supported
            }
            break;

            // Set Speed
        case 0x0F:
            if (eparm < 0x20)
                m_nSpeed = note->effect_parms;
            else
                m_nBPM = note->effect_parms;
            break;

        default:
            break;
        }

        // If we have something playing then set the frequency
        if (m_TrackDat[track].period > 0)
            m_TrackDat[track].freq = Period2Freq(m_TrackDat[track].period);
    }

    // Update our row and orders
    m_nRow = newrow;
    m_nOrder = neworder;
}

static bool MixSubChunk(short *buffer, int numsamples)
{
    int i, track;
    // Setup left and right channels
    short *left;
    short *right;
    left = buffer;
    right = buffer + 1;
    // Set up a mixing buffer and clear it
    for (i = 0; i < numsamples * 2; i++)
        buffer[i] = 0;

    // Loop through each channel and process note data
    for (track = 0; track < m_nNumTracks; track++) {
        // Make sure I'm actually playing something
        if (m_TrackDat[track].sample <= 0)
            continue;

        // Make sure this sample actually contains sound data
        if (!m_Samples[m_TrackDat[track].sample].data_length)
            continue;
        {
            // Set up for the mix loop
            short *mixed = ((track & 3) == 0)
                || ((track & 3) == 3) ? left : right;
            unsigned char *sample = (unsigned char *) &m_Samples[m_TrackDat[track].sample].data[0];
            int nLength = m_Samples[m_TrackDat[track].sample].nLength << FRAC_BITS;
            int nLoopLength = m_Samples[m_TrackDat[track].sample].nLoopLength << FRAC_BITS;
            int nLoopEnd = m_Samples[m_TrackDat[track].sample].nLoopEnd << FRAC_BITS;
            int finetune = m_Samples[m_TrackDat[track].sample].nFineTune;
            int notenum = m_TrackDat[track].period_index + finetune;
            float freq = m_TrackDat[track].freq;
            int pos = m_TrackDat[track].pos;
            int deltapos = (int) (freq * (1 << FRAC_BITS) / PLAYBACK_FREQ);
            int *VolumeTablePtr = VolumeTable[m_TrackDat[track].mixvol];
            int mixpos = 0;
            int samples_to_mix = numsamples;
            notenum = notenum < 0 ? 0 : notenum >= NUMNOTES ? NUMNOTES - 1 : notenum;

            while (samples_to_mix) {
                int thiscount;

                // If I'm a looping sample then I need to check if it's time to
                // loop back. I also need to figure out how many samples I can mix
                // before I need to loop again
                if (nLoopEnd > (2 << FRAC_BITS)) {
                    if (pos >= nLoopEnd)
                        pos -= nLoopLength;
                    thiscount = min(samples_to_mix, (nLoopEnd - pos - 1) / deltapos + 1);
                    // above returns the smaller parameter

                    samples_to_mix -= thiscount;
                }
                // If I'm not a looping sample then mix until I'm done playing
                // the entire sample
                else {
                    // If we've already reached the end of the sample then forget it
                    if (pos >= nLength)
                        thiscount = 0;
                    else
                        thiscount = min(numsamples, (nLength - pos - 1) / deltapos + 1);
                    samples_to_mix = 0;
                }

                // Inner Loop start
                for (i = 0; i < thiscount; i++) {
                    // Mix this sample in and update our position
#ifdef OVERSAMPLE
                    //  This smooths the sound a bit by oversampling, but
                    //  uses up more cpu (only a bit)
                    int sample1 = VolumeTablePtr[sample[pos >> FRAC_BITS]];
                    int sample2 = VolumeTablePtr[sample[(pos >> FRAC_BITS) + 1]];
                    int frac1 = pos & ((1 << FRAC_BITS) - 1);
                    int frac2 = (1 << FRAC_BITS) - frac1;
                    mixed[mixpos] += ((sample1 * frac2) + (sample2 * frac1)) >> FRAC_BITS;
                    mixpos += 2;
#else
                    //  This is normal plain mixing
                    mixed[mixpos] += VolumeTablePtr[sample[pos >> FRAC_BITS]];
                    mixpos += 2;
#endif
                    pos += deltapos;
                }
                // Inner Loop end
            }
            // Save current position
            m_TrackDat[track].pos = pos;
        }
    }

    return true;
}

static void UpdateEffects()
{
    int track;
    // Loop through each channel
    for (track = 0; track < m_nNumTracks; track++) {
        // Get note data
        NoteData *note = &m_CurrentRow->note[track];

        // Parse it
        int effect = note->effect;	// grab the effect number
        int eparm = note->effect_parms;	// grab the effect parameter
        int eparmx = eparm >> 4;	// grab the effect parameter x
        int eparmy = eparm & 0xF;	// grab the effect parameter y

        // Process it
        switch (effect) {
            // Arpeggio
        case 0x00:
            if (eparm > 0) {
                int notenum, period;
                switch (m_nTick % 3) {
                case 0:
                    period = m_TrackDat[track].period;
                    break;
                case 1:
                    notenum =
                        m_TrackDat[track].period_index + 8 * eparmx + m_Samples[m_TrackDat[track].sample].nFineTune;
                    period = PeriodTable[notenum];
                    break;
                case 2:
                    notenum =
                        m_TrackDat[track].period_index + 8 * eparmy + m_Samples[m_TrackDat[track].sample].nFineTune;
                    period = PeriodTable[notenum];
                    break;
                default:
                    period = 0;
                    break;
                }
                m_TrackDat[track].freq = Period2Freq(period);
            }
            break;


            // Porta up
        case 0x01:
            m_TrackDat[track].period -= eparm;	// subtract freq
            if (m_TrackDat[track].period < 54)
                m_TrackDat[track].period = 54;	// stop at C-5
            m_TrackDat[track].freq = Period2Freq(m_TrackDat[track].period);
            break;

            // Porta down
        case 0x02:
            m_TrackDat[track].period += eparm;	// add freq
            m_TrackDat[track].freq = Period2Freq(m_TrackDat[track].period);
            break;

            // Porta to note
        case 0x03:
            DoPorta(track);
            break;

            // Vibrato
        case 0x04:
            DoVibrato(track);
            break;

            // Porta + Vol Slide
        case 0x05:
            DoPorta(track);
            SlideVolume(track, eparmx - eparmy);
            m_TrackDat[track].mixvol = m_TrackDat[track].volume;
            break;

            // Vibrato + Vol Slide
        case 0x06:
            DoVibrato(track);
            SlideVolume(track, eparmx - eparmy);
            m_TrackDat[track].mixvol = m_TrackDat[track].volume;
            break;

            // Tremolo
        case 0x07:
            DoTremolo(track);
            break;

            // Pan
        case 0x08:
            break;		// note effect

            // Sample offset
        case 0x09:
            break;		// note effect

            // Volume slide
        case 0x0A:
            SlideVolume(track, eparmx - eparmy);
            m_TrackDat[track].mixvol = m_TrackDat[track].volume;
            break;

            // Jump To Pattern
        case 0x0B:
            break;		// note effect

            // Set Volume
        case 0x0C:
            break;		// note effect

            // Pattern Break
        case 0x0D:
            break;		// note effect

            // Extended effects
        case 0x0E:
            switch (eparmx) {
                // Retrig note
            case 0x9:
                break;		// not supported

                // Cut note
            case 0xC:
                if (m_nTick == eparmy) {
                    m_TrackDat[track].volume = 0;
                    m_TrackDat[track].mixvol = m_TrackDat[track].volume;
                }
                break;

                // Delay note
            case 0xD:
                break;		// not supported

                // All other Exy effects are note effects
            }

            // Set Speed
        case 0x0F:
            break;		// note effect
        }
    }
}

static void SlideVolume(int track, int amount)
{
    amount += m_TrackDat[track].volume;
    m_TrackDat[track].volume = amount < 0 ? 0 : amount > 64 ? 64 : amount;
}

static void DoPorta(int track)
{
    if (m_TrackDat[track].period < m_TrackDat[track].porta) {
        m_TrackDat[track].period += m_TrackDat[track].portasp;
        if (m_TrackDat[track].period > m_TrackDat[track].porta)
            m_TrackDat[track].period = m_TrackDat[track].porta;
    } else if (m_TrackDat[track].period > m_TrackDat[track].porta) {
        m_TrackDat[track].period -= m_TrackDat[track].portasp;
        if (m_TrackDat[track].period < m_TrackDat[track].porta)
            m_TrackDat[track].period = m_TrackDat[track].porta;
    }
    m_TrackDat[track].freq = Period2Freq(m_TrackDat[track].period);
}

static void DoVibrato(int track)
{
    int vib = m_TrackDat[track].vibdep * sintab[m_TrackDat[track].sinepos] >> 7;	// div 128
    if (m_TrackDat[track].sineneg == 0)
        m_TrackDat[track].freq = Period2Freq(m_TrackDat[track].period + vib);
    else
        m_TrackDat[track].freq = Period2Freq(m_TrackDat[track].period - vib);
    m_TrackDat[track].sinepos += m_TrackDat[track].vibspe;
    if (m_TrackDat[track].sinepos > 31) {
        m_TrackDat[track].sinepos -= 32;
        m_TrackDat[track].sineneg = ~m_TrackDat[track].sineneg;	// flip pos/neg flag
    }
}

static void DoTremolo(int track)
{
    int vib = m_TrackDat[track].tremdep * sintab[m_TrackDat[track].sinepos] >> 6;	// div64
    if (m_TrackDat[track].sineneg == 0) {
        if (m_TrackDat[track].volume + vib > 64)
            vib = 64 - m_TrackDat[track].volume;
        m_TrackDat[track].mixvol = m_TrackDat[track].volume + vib;
    } else {
        if (m_TrackDat[track].volume - vib < 0)
            vib = m_TrackDat[track].volume;
        m_TrackDat[track].mixvol = m_TrackDat[track].volume + vib;
    }

    m_TrackDat[track].sinepos += m_TrackDat[track].tremspe;
    if (m_TrackDat[track].sinepos > 31) {
        m_TrackDat[track].sinepos -= 32;
        m_TrackDat[track].sineneg = ~m_TrackDat[track].sineneg;	// flip pos/neg flag
    }
}

// Sets the master volume for the mod being played.
// The value should be from 0 (silence) to 64 (max volume)
static void SetMasterVolume(int volume)
{
    int i, j;
    for (i = 0; i < 65; i++)
        for (j = 0; j < 256; j++)
            VolumeTable[i][j] = volume * i * (int) (char) j / 64;
}

// 16-bit word values in a mod are stored in the Motorola
// Most-Significant-Byte-First format.
// They're also stored at half their actual value, thus doubling their range.
// This function accepts a pointer to such a word and returns it's integer value
// NOTE: relic from pc testing.
static int ReadModWord(unsigned char *buffer, int pos)
{
    int byte1 = (int) (unsigned char) *(buffer + pos);
    int byte2 = (int) (unsigned char) *(buffer + pos + 1);
    return ((byte1 * 256) + byte2) * 2;
}
